# stdlib
from datetime import datetime
from datetime import timedelta
import logging

# third party
from pydantic import EmailStr

# relative
from ...abstract_server import AbstractServer
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from ..context import AuthedServiceContext
from ..notification.email_templates import PasswordResetTemplate
from ..notification.notifications import Notification
from ..response import SyftSuccess
from ..service import AbstractService
from .notifier import EmailFrequency
from .notifier import NotificationPreferences
from .notifier import NotifierSettings
from .notifier import UserNotificationActivity
from .notifier_enums import EMAIL_TYPES
from .notifier_enums import NOTIFICATION_FREQUENCY
from .notifier_enums import NOTIFIERS
from .notifier_stash import NotifierStash

logger = logging.getLogger(__name__)


class RateLimitException(SyftException):
    public_message = "Rate limit exceeded."


@serializable(canonical_name="NotifierService", version=1)
class NotifierService(AbstractService):
    stash: NotifierStash

    def __init__(self, store: DBManager) -> None:
        self.stash = NotifierStash(store=store)

    @as_result(StashException)
    def settings(
        self,
        context: AuthedServiceContext,
    ) -> NotifierSettings:
        """Get Notifier Settings

        Args:
            context: The request context
        Returns:
            NotifierSettings | None: The notifier settings, if it exists; None otherwise.
        """
        return self.stash.get(credentials=context.credentials).unwrap(
            public_message="Error getting notifier settings"
        )

    def user_settings(
        self,
        context: AuthedServiceContext,
    ) -> NotificationPreferences:
        user_view = context.server.services.user.get_current_user(context)
        notifications = user_view.notifications_enabled
        return NotificationPreferences(
            email=notifications[NOTIFIERS.EMAIL],
            sms=notifications[NOTIFIERS.SMS],
            slack=notifications[NOTIFIERS.SLACK],
            app=notifications[NOTIFIERS.APP],
        )

    def _set_notifier(self, context: AuthedServiceContext, active: bool) -> SyftSuccess:
        notifier = self.stash.get(credentials=context.credentials).unwrap(
            public_message="Notifier settings not found."
        )
        notifier.active = active
        self.stash.update(credentials=context.credentials, obj=notifier).unwrap()

        active_s = "active" if active else "inactive"
        return SyftSuccess(message=f"Notifier set to {active_s}")

    def set_notifier_active_to_false(
        self, context: AuthedServiceContext
    ) -> SyftSuccess:
        """
        Essentially a duplicate of turn_off method.
        """
        notifier = self.stash.get(credentials=context.credentials).unwrap()
        notifier.active = False
        self.stash.update(credentials=context.credentials, obj=notifier).unwrap()
        return SyftSuccess(message="notifier.active set to false.")

    @as_result(SyftException)
    def turn_on(
        self,
        context: AuthedServiceContext,
        email_username: str | None = None,
        email_password: str | None = None,
        email_sender: str | None = None,
        email_server: str | None = None,
        email_port: int | None = 587,
    ) -> SyftSuccess:
        """Turn on email notifications.

        Args:
            email_username (Optional[str]): Email server username. Defaults to None.
            email_password (Optional[str]): Email email server password. Defaults to None.
            sender_email (Optional[str]): Email sender email. Defaults to None.
        Returns:
            SyftSuccess: success response.

        Raises:
            SyftException: any error that occurs during the process
        """

        # 1 -  If something went wrong at db level, return the error
        notifier = self.stash.get(credentials=context.credentials).unwrap()

        # 2 - If one of the credentials are set alone, return an error
        if (email_username and not email_password) or (
            not email_username and email_password
        ):
            raise SyftException(
                public_message="You must provide both username and password"
            )

        # 3 - If notifier doesn't have a email server / port and the user didn't provide them, return an error
        if not (email_server and email_port) and not notifier.email_server:
            raise SyftException(
                public_message="You must provide both server and port to enable notifications."
            )

        logger.debug("Got notifier from db")
        skip_auth: bool = False
        # If no new credentials provided, check for existing ones
        if not (email_username and email_password):
            if not (notifier.email_username and notifier.email_password):
                skip_auth = True
            else:
                logger.debug("No new credentials provided. Using existing ones.")
                email_password = notifier.email_password
                email_username = notifier.email_username

        valid_credentials = True
        if not skip_auth:
            valid_credentials = notifier.validate_email_credentials(
                username=email_username,
                password=email_password,
                server=email_server or notifier.email_server,
                port=email_port or notifier.email_port,
            )

        if not valid_credentials:
            logger.error("Invalid SMTP credentials.")
            raise SyftException(public_message=("Invalid SMTP credentials."))

        notifier.email_password = email_password
        notifier.email_username = email_username

        if email_server:
            notifier.email_server = email_server
        if email_port:
            notifier.email_port = email_port

        # Email sender verification
        if not email_sender and not notifier.email_sender:
            raise SyftException(
                public_message="You must provide a sender email address to enable notifications."
            )

        # If email_rate_limit isn't defined yet.
        if not notifier.email_rate_limit:
            notifier.email_rate_limit = {PasswordResetTemplate.__name__: 3}

        if email_sender:
            try:
                EmailStr._validate(email_sender)
            except ValueError:
                raise SyftException(
                    publiccmessage="Invalid sender email address. Please check your email address."
                )
            notifier.email_sender = email_sender

        notifier.active = True
        logger.debug(
            "Email credentials are valid. Updating the notifier settings in the db."
        )

        self.stash.update(credentials=context.credentials, obj=notifier).unwrap()
        context.server.services.settings.update(context, notifications_enabled=True)
        return SyftSuccess(message="Notifications enabled successfully.")

    @as_result(SyftException)
    def set_email_batch(
        self,
        context: AuthedServiceContext,
        email_type: EMAIL_TYPES,
        frequency: NOTIFICATION_FREQUENCY,
        start_time: str = "",
    ) -> SyftSuccess:
        if start_time == "" and frequency is not NOTIFICATION_FREQUENCY.INSTANT:
            raise SyftException(
                "If frequency isn't INSTANT, you must set a start time for the notifications to be dispatched."
            )

        if frequency is not NOTIFICATION_FREQUENCY.INSTANT:
            start_time = start_time.lower()
            try:
                if "pm" in start_time or "am" in start_time:
                    time_obj = datetime.strptime(start_time, "%I:%M %p")
                else:
                    time_obj = datetime.strptime(start_time, "%H:%M")
            except ValueError:
                raise SyftException(
                    "Invalid time format."
                    + "Please enter the start time in one of the following format examples:"
                    + "'14:00' or '2:00 PM'."
                )
        else:
            time_obj = datetime.now()

        notifier = self.stash.get(credentials=context.credentials).unwrap()
        notifier.email_frequency[email_type.value] = EmailFrequency(
            frequency=frequency, start_time=time_obj
        )
        self.stash.update(credentials=context.credentials, obj=notifier).unwrap()
        return SyftSuccess(message="Configuration set successfully.")

    @as_result(StashException)
    def turn_off(
        self,
        context: AuthedServiceContext,
    ) -> SyftSuccess:
        """
        Turn off email notifications service.
        PySyft notifications will still work.
        """
        notifier = self.stash.get(credentials=context.credentials).unwrap()

        notifier.active = False
        self.stash.update(credentials=context.credentials, obj=notifier).unwrap()
        context.server.services.settings.update(context, notifications_enabled=False)
        return SyftSuccess(message="Notifications disabled succesfullly")

    @as_result(SyftException)
    def activate(
        self, context: AuthedServiceContext, notifier_type: NOTIFIERS = NOTIFIERS.EMAIL
    ) -> SyftSuccess:
        """
        Activate email notifications for the authenticated user.
        This will only work if the datasite owner has enabled notifications.
        """
        return context.server.services.user.enable_notifications(
            context, notifier_type=notifier_type
        ).unwrap()

    @as_result(SyftException)
    def deactivate(
        self, context: AuthedServiceContext, notifier_type: NOTIFIERS = NOTIFIERS.EMAIL
    ) -> SyftSuccess:
        """Deactivate email notifications for the authenticated user
        This will only work if the datasite owner has enabled notifications.
        """
        result = context.server.services.user.disable_notifications(
            context, notifier_type=notifier_type
        )
        return result

    def is_time_to_dispatch(
        self, notification_frequency: EmailFrequency, current_time: datetime
    ) -> bool:
        frequency = notification_frequency.frequency
        start_time = notification_frequency.start_time

        # Define period_timedelta based on frequency
        if frequency == NOTIFICATION_FREQUENCY.INSTANT:
            return True
        if frequency == NOTIFICATION_FREQUENCY.SIX_HOURS:
            period = timedelta(hours=6)
        elif frequency == NOTIFICATION_FREQUENCY.TWELVE_HOURS:
            period = timedelta(hours=12)
        elif frequency == NOTIFICATION_FREQUENCY.DAILY:
            period = timedelta(days=1)
        elif frequency == NOTIFICATION_FREQUENCY.WEEKLY:
            period = timedelta(weeks=1)

        # Calculate how many full periods have passed since start_time
        elapsed_time = current_time - start_time
        if elapsed_time < timedelta(0):
            return False  # Current time is before the start time

        periods_elapsed = int(elapsed_time // period)
        next_dispatch_time = start_time + periods_elapsed * period

        # Allow a small margin of error (e.g., 1 minute) to account for processing delays
        margin = timedelta(minutes=1)
        time_difference = current_time - next_dispatch_time

        # Check if current_time is within the margin of the scheduled dispatch time
        return timedelta(0) <= time_difference <= margin

    @staticmethod
    @as_result(SyftException)
    def init_notifier(
        server: AbstractServer,
        email_username: str | None = None,
        email_password: str | None = None,
        email_sender: str | None = None,
        smtp_port: int | None = None,
        smtp_host: str | None = None,
    ) -> SyftSuccess | None:
        """Initialize Notifier settings for a Server.
        If settings already exist, it will use the existing one.
        If not, it will create a new one.

        Args:
            server: Server to initialize the notifier
            active: If notifier should be active
            email_username: Email username to send notifications
            email_password: Email password to send notifications
        Raises:
            Exception: If something went wrong
        Returns:
            SyftSuccess
        """
        try:
            # Create a new NotifierStash since its a static method.
            notifier_stash = NotifierStash(store=server.db)
            should_update = False

            # Get the notifier
            # If notifier doesn't exist, create a new one
            try:
                notifier = notifier_stash.get(server.signing_key.verify_key).unwrap()
                should_update = True
            except NotFoundException:
                notifier = NotifierSettings()
                notifier.active = False  # Default to False

            # TODO: this should be a method in NotifierSettings
            if email_username and email_password:
                validation_result = notifier.validate_email_credentials(
                    username=email_username,
                    password=email_password,
                    server=smtp_host,
                    port=smtp_port,
                )

                sender_not_set = not email_sender and not notifier.email_sender

                if not validation_result or sender_not_set:
                    logger.error("Notifier validation error")
                    notifier.active = False
                else:
                    notifier.email_password = email_password
                    notifier.email_username = email_username
                    notifier.email_sender = email_sender
                    notifier.email_server = smtp_host
                    notifier.email_port = smtp_port
                    # Default daily email rate limit per user
                    notifier.email_rate_limit = {PasswordResetTemplate.__name__: 3}
                    notifier.active = True

            if should_update:
                notifier_stash.update(
                    credentials=server.signing_key.verify_key, obj=notifier
                ).unwrap()
            else:
                notifier_stash.set(server.signing_key.verify_key, notifier).unwrap()
            return SyftSuccess(
                message="Notifier initialized successfully", value=notifier
            )
        except Exception as e:
            raise SyftException.from_exception(
                e, public_message=f"Error initializing notifier. {e}"
            )

    def set_email_rate_limit(
        self, context: AuthedServiceContext, email_type: EMAIL_TYPES, daily_limit: int
    ) -> SyftSuccess:
        notifier = self.stash.get(context.credentials).unwrap(
            public_message="Couldn't set the email rate limit."
        )
        notifier.email_rate_limit[email_type.value] = daily_limit
        self.stash.update(credentials=context.credentials, obj=notifier)

        return SyftSuccess(message="Email rate limit updated!")

    # This is not a public API.
    # This method is used by other services to dispatch notifications internally
    @as_result(SyftException, RateLimitException)
    def dispatch_notification(
        self, context: AuthedServiceContext, notification: Notification
    ) -> SyftSuccess:
        admin_key = context.server.services.user.root_verify_key

        # Silently fail on notification not delivered
        try:
            notifier = self.stash.get(admin_key).unwrap(
                public_message="The mail service ran out of quota or some notifications failed to be delivered.\n"
                + "Please check the health of the mailing server."
            )
        except NotFoundException:
            logger.debug("There is no notifier service to ship the notification")
            raise SyftException(
                public_message="No notifier service to ship the notification."
            )
        except StashException as exc:
            logger.error(f"Error getting notifier settings: {exc}")
            raise SyftException(
                public_message="Failed to get notifier settings. Please check the logs."
            )

        # If notifier is active
        if notifier.active and notification.email_template is not None:
            logger.debug("Checking user email activity")

            if notifier.email_activity.get(notification.email_template.__name__, None):
                user_activity = notifier.email_activity[
                    notification.email_template.__name__
                ].get(notification.to_user_verify_key, None)

                # If there's no user activity
                if user_activity is None:
                    notifier.email_activity[notification.email_template.__name__][
                        notification.to_user_verify_key
                    ] = UserNotificationActivity(count=1, date=datetime.now())
                else:  # If there's a previous user activity
                    current_state: UserNotificationActivity = notifier.email_activity[
                        notification.email_template.__name__
                    ][notification.to_user_verify_key]
                    date_refresh = abs(datetime.now() - current_state.date).days > 1

                    limit = notifier.email_rate_limit.get(
                        notification.email_template.__name__, 0
                    )
                    still_in_limit = current_state.count < limit
                    # Time interval reseted.
                    if date_refresh:
                        current_state.count = 1
                        current_state.date = datetime.now()
                    # Time interval didn't reset yet.
                    elif still_in_limit or not limit:
                        current_state.count += 1
                        current_state.date = datetime.now()
                    else:
                        raise RateLimitException(
                            public_message="Couldn't send the email. You have surpassed the"
                            + " email threshold limit. Please try again later."
                        )
            else:
                notifier.email_activity[notification.email_template.__name__] = {
                    notification.to_user_verify_key: UserNotificationActivity(
                        count=1, date=datetime.now()
                    )
                }

            email_frequency = notifier.email_frequency.get(
                notification.email_template.__name__,
                EmailFrequency(frequency=NOTIFICATION_FREQUENCY.INSTANT),
            )

            if email_frequency.frequency == NOTIFICATION_FREQUENCY.INSTANT:
                notifier.send_notifications(
                    context=context, notification=notification
                ).unwrap()
            else:
                queue_dict = notifier.email_queue.get(
                    notification.email_template.__name__, {}
                )
                if len(queue_dict) == 0:
                    notifier.email_queue[notification.email_template.__name__] = (
                        queue_dict
                    )

                user_queue = queue_dict.get(notification.to_user_verify_key, [])

                if len(user_queue) == 0:
                    queue_dict[notification.to_user_verify_key] = user_queue

                user_queue.append(notification)

            self.stash.update(credentials=admin_key, obj=notifier).unwrap()

        # If notifier isn't active, return None
        return SyftSuccess(message="Notification dispatched successfully")



# third party
from sqlalchemy import URL

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.uid import UID
from .db import DBManager
from .sqlite import DBConfig


@serializable(canonical_name="PostgresDBConfig", version=1)
class PostgresDBConfig(DBConfig):
    host: str
    port: int
    user: str
    password: str
    database: str

    @property
    def connection_string(self) -> str:
        return URL.create(
            "postgresql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
        ).render_as_string(hide_password=False)


class PostgresDBManager(DBManager[PostgresDBConfig]):
    def update_settings(self) -> None:
        return super().update_settings()

    @classmethod
    def random(
        cls: type,
        *,
        config: PostgresDBConfig,
        server_uid: UID | None = None,
        root_verify_key: SyftVerifyKey | None = None,
    ) -> "PostgresDBManager":
        root_verify_key = root_verify_key or SyftVerifyKey.generate()
        server_uid = server_uid or UID()
        return PostgresDBManager(
            config=config, server_uid=server_uid, root_verify_key=root_verify_key
        )



# stdlib
from collections import defaultdict
import logging
from typing import Any

# relative
from ...client.api import ServerIdentity
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..action.action_permissions import StoragePermission
from ..api.api import TwinAPIEndpoint
from ..code.user_code import UserCodeStatusCollection
from ..context import AuthedServiceContext
from ..job.job_stash import Job
from ..log.log import SyftLog
from ..request.request import Request
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from .sync_stash import SyncStash
from .sync_state import SyncState

logger = logging.getLogger(__name__)


def get_store(context: AuthedServiceContext, item: SyncableSyftObject) -> ObjectStash:
    return get_store_by_type(context=context, obj_type=type(item))


def get_store_by_type(context: AuthedServiceContext, obj_type: type) -> ObjectStash:
    if issubclass(obj_type, ActionObject):
        service = context.server.services.action  # type: ignore
        return service.stash  # type: ignore
    service = context.server.get_service(TYPE_TO_SERVICE[obj_type])  # type: ignore
    return service.stash


@instrument
@serializable(canonical_name="SyncService", version=1)
class SyncService(AbstractService):
    stash: SyncStash

    def __init__(self, store: DBManager):
        self.stash = SyncStash(store=store)

    def add_actionobject_read_permissions(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject,
        new_permissions: list[ActionObjectPermission],
    ) -> None:
        action_stash = context.server.services.action.stash
        for permission in new_permissions:
            if permission.permission == ActionPermission.READ:
                action_stash.add_permission(permission)

        blob_id = action_object.syft_blob_storage_entry_id
        if blob_id:
            blob_stash = context.server.services.blob_storage.stash
            for permission in new_permissions:
                if permission.permission == ActionPermission.READ:
                    permission_blob = ActionObjectPermission(
                        uid=blob_id,
                        permission=permission.permission,
                        credentials=permission.credentials,
                    )
                    blob_stash.add_permission(permission_blob)

    def set_obj_ids(self, context: AuthedServiceContext, x: Any) -> None:
        if hasattr(x, "__dict__") and isinstance(x, SyftObject):
            for val in x.__dict__.values():
                if isinstance(val, list | tuple):
                    for v in val:
                        self.set_obj_ids(context, v)
                elif isinstance(val, dict):
                    for v in val.values():
                        self.set_obj_ids(context, v)
                else:
                    self.set_obj_ids(context, val)
            x.syft_server_location = context.server.id  # type: ignore
            x.syft_client_verify_key = context.credentials
            if hasattr(x, "server_uid"):
                x.server_uid = context.server.id  # type: ignore

    def transform_item(
        self,
        context: AuthedServiceContext,
        item: SyncableSyftObject,
    ) -> SyftObject:
        if isinstance(item, UserCodeStatusCollection):
            identity = ServerIdentity.from_server(context.server)
            res = {}
            for approval_decision in item.status_dict.values():
                # todo, check if they are actually only two servers
                res[identity] = approval_decision
            item.status_dict = res

        self.set_obj_ids(context, item)
        return item

    @as_result(ValueError)
    def get_stash_for_item(
        self, context: AuthedServiceContext, item: SyftObject
    ) -> ObjectStash:
        services = list(context.server.service_path_map.values())  # type: ignore

        all_stashes = {}
        for serv in services:
            if (_stash := getattr(serv, "stash", None)) is not None:
                all_stashes[_stash.object_type] = _stash

        stash = all_stashes.get(type(item), None)
        if stash is None:
            raise ValueError(f"Could not find stash for {type(item)}")
        return stash

    def add_permissions_for_item(
        self,
        context: AuthedServiceContext,
        item: SyftObject,
        new_permissions: list[ActionObjectPermission],
    ) -> None:
        if isinstance(item, ActionObject):
            raise ValueError("ActionObject permissions should be added separately")
        else:
            store = get_store(context, item)  # type: ignore
            for permission in new_permissions:
                if permission.permission == ActionPermission.READ:
                    store.add_permission(permission)

    def add_storage_permissions_for_item(
        self,
        context: AuthedServiceContext,
        item: SyftObject,
        new_permissions: list[StoragePermission],
    ) -> None:
        store = get_store(context, item)
        store.add_storage_permissions(new_permissions)

    @as_result(SyftException)
    def set_object(
        self, context: AuthedServiceContext, item: SyncableSyftObject
    ) -> SyftObject:
        stash = self.get_stash_for_item(context, item).unwrap()
        creds = context.credentials

        obj = None
        try:
            obj = stash.get_by_uid(context.credentials, item.id).unwrap()
        except (SyftException, KeyError):
            obj = None

        exists = obj is not None

        if isinstance(item, TwinAPIEndpoint):
            # we need the side effect of set function
            # to create an action object
            res = context.server.services.api.set(context=context, endpoint=item)
            return item

        if exists:
            res = stash.update(creds, item).unwrap()
        else:
            # Storage permissions are added separately
            res = stash.set(
                creds,
                item,
                add_storage_permission=False,
            ).unwrap()

        return res

    @service_method(
        path="sync.sync_items",
        name="sync_items",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def sync_items(
        self,
        context: AuthedServiceContext,
        items: list[SyncableSyftObject],
        permissions: dict[type, list[ActionObjectPermission]],
        storage_permissions: list[StoragePermission],
        ignored_batches: dict[UID, int],
        unignored_batches: set[UID],
    ) -> SyftSuccess:
        permissions_dict = defaultdict(list)
        for permission_list in permissions.values():
            for permission in permission_list:
                permissions_dict[permission.uid].append(permission)

        item_ids = [item.id.id for item in items]

        # If we just want to add permissions without having an object
        # This should happen only for the high side when we sync results but
        # we need to add permissions for the DS to properly show the status of the requests
        for obj_type, permission_list in permissions.items():
            for permission in permission_list:
                if permission.uid in item_ids:
                    continue
                if obj_type not in [Job, SyftLog, Request] and not issubclass(
                    obj_type, ActionObject
                ):
                    raise SyftException(
                        public_message="Permission for object type not supported!"
                    )
                if issubclass(obj_type, ActionObject):
                    store = context.server.services.action.stash
                else:
                    service = context.server.get_service(TYPE_TO_SERVICE[obj_type])
                    store = service.stash  # type: ignore[assignment]
                if permission.permission == ActionPermission.READ:
                    store.add_permission(permission)

        storage_permissions_dict = defaultdict(list)
        for storage_permission in storage_permissions:
            storage_permissions_dict[storage_permission.uid].append(storage_permission)

        for item in items:
            new_permissions = permissions_dict[item.id.id]
            new_storage_permissions = storage_permissions_dict[item.id.id]
            if isinstance(item, ActionObject):
                self.add_actionobject_read_permissions(context, item, new_permissions)
                self.add_storage_permissions_for_item(
                    context, item, new_storage_permissions
                )
            else:
                item = self.transform_item(context, item)  # type: ignore[unreachable]
                self.set_object(context, item).unwrap()
                self.add_permissions_for_item(context, item, new_permissions)
                self.add_storage_permissions_for_item(
                    context, item, new_storage_permissions
                )

        # NOTE include_items=False to avoid snapshotting the database
        # Snapshotting is disabled to avoid mongo size limit and performance issues
        new_state = self.build_current_state(
            context,
            new_items=items,
            new_ignored_batches=ignored_batches,
            new_unignored_batches=unignored_batches,
            include_items=False,
        ).unwrap()

        self.stash.set(context.credentials, new_state).unwrap()

        message = f"Synced {len(items)} items"
        if len(ignored_batches) > 0:
            message += f", ignored {len(ignored_batches)} batches"
        if len(unignored_batches) > 0:
            message += f", unignored {len(unignored_batches)} batches"
        return SyftSuccess(message=message)

    @service_method(
        path="sync.get_permissions",
        name="get_permissions",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_permissions(
        self,
        context: AuthedServiceContext,
        items: list[SyncableSyftObject],
    ) -> tuple[dict[UID, set[str]], dict[UID, set[UID]]]:
        permissions: dict[UID, set[str]] = {}
        storage_permissions: dict[UID, set[UID]] = {}

        for item in items:
            store = get_store(context, item)
            if store is not None:
                # TODO fix error handling
                uid = item.id.id
                permissions[uid] = store._get_permissions_for_uid(uid).unwrap()

                # TODO fix error handling for storage permissions
                storage_permissions[uid] = store._get_storage_permissions_for_uid(
                    uid
                ).unwrap()
        return permissions, storage_permissions

    @as_result(SyftException)
    def _get_all_items_for_jobs(
        self,
        context: AuthedServiceContext,
    ) -> tuple[list[SyncableSyftObject], dict[UID, str]]:
        """
        Returns all Jobs, along with their Logs, ExecutionOutputs and ActionObjects
        """
        items_for_jobs: list[SyncableSyftObject] = []
        errors = {}
        jobs = context.server.services.job.get_all(context)

        for job in jobs:
            try:
                job_items = self._get_job_batch(context, job).unwrap()
                items_for_jobs.extend(job_items)
            except SyftException as exc:
                logger.info(
                    f"Job {job.id} could not be added to SyncState: {exc._private_message or exc.public_message}"
                )
                errors[job.id] = str(exc)

        return (items_for_jobs, errors)

    @as_result(SyftException)
    def _get_job_batch(
        self, context: AuthedServiceContext, job: Job
    ) -> list[SyncableSyftObject]:
        job_batch = [job]

        log = context.server.services.log.get(context, job.log_id)
        job_batch.append(log)

        try:
            output = context.server.services.output.get_by_job_id(context, job.id)
        except NotFoundException:
            output = None

        if output is not None:
            job_batch.append(output)
            job_result_ids = set(output.output_id_list)
        else:
            job_result_ids = set()

        if isinstance(job.result, ActionObject):
            job_result_ids.add(job.result.id.id)

        for result_id in job_result_ids:
            # TODO: unwrap
            action_object = context.server.services.action.get(context, result_id)
            job_batch.append(action_object)

        return job_batch

    @as_result(SyftException)
    def get_all_syncable_items(
        self, context: AuthedServiceContext
    ) -> tuple[list[SyncableSyftObject], dict[UID, str]]:
        all_items: list[SyncableSyftObject] = []

        # NOTE Jobs are handled separately
        services_to_sync = [
            "requestservice",
            "usercodeservice",
            "usercodestatusservice",
            "apiservice",
        ]

        for service_name in services_to_sync:
            service = context.server.get_service(service_name)
            items = service.get_all(context)
            all_items.extend(items)

        # Gather jobs, logs, outputs and action objects
        items_for_jobs, errors = self._get_all_items_for_jobs(context).unwrap()
        # items_for_jobs, errors = items_for_jobs
        all_items.extend(items_for_jobs)

        return (all_items, errors)

    @as_result(SyftException)
    def build_current_state(
        self,
        context: AuthedServiceContext,
        new_items: list[SyncableSyftObject] | None = None,
        new_ignored_batches: dict[UID, int] | None = None,
        new_unignored_batches: set[UID] | None = None,
        include_items: bool = True,
    ) -> SyncState:
        new_items = new_items if new_items is not None else []
        new_ignored_batches = (
            new_ignored_batches if new_ignored_batches is not None else {}
        )
        unignored_batches: set[UID] = (
            new_unignored_batches if new_unignored_batches is not None else set()
        )
        if include_items:
            objects, errors = self.get_all_syncable_items(context).unwrap()
            permissions, storage_permissions = self.get_permissions(context, objects)
        else:
            objects = []
            errors = {}
            permissions = {}
            storage_permissions = {}

        try:
            previous_state = self.stash.get_latest(
                credentials=context.credentials
            ).unwrap()
        except NotFoundException:
            previous_state = None

        if previous_state is not None:
            previous_state_link = LinkedObject.from_obj(
                obj=previous_state,
                service_type=SyncService,
                server_uid=context.server.id,  # type: ignore
            )
            previous_ignored_batches = previous_state.ignored_batches
        else:
            previous_state_link = None
            previous_ignored_batches = {}

        ignored_batches = {
            **previous_ignored_batches,
            **new_ignored_batches,
        }

        ignored_batches = {
            k: v for k, v in ignored_batches.items() if k not in unignored_batches
        }

        object_sync_dates = (
            previous_state.object_sync_dates.copy() if previous_state else {}
        )
        for obj in new_items:
            object_sync_dates[obj.id.id] = DateTime.now()

        new_state = SyncState(
            server_uid=context.server.id,  # type: ignore
            server_name=context.server.name,  # type: ignore
            server_side_type=context.server.server_side_type,  # type: ignore
            previous_state_link=previous_state_link,
            permissions=permissions,
            storage_permissions=storage_permissions,
            ignored_batches=ignored_batches,
            object_sync_dates=object_sync_dates,
            errors=errors,
        )

        new_state.add_objects(objects, context)

        return new_state

    @service_method(
        path="sync._get_state",
        name="_get_state",
        roles=ADMIN_ROLE_LEVEL,
    )
    def _get_state(self, context: AuthedServiceContext) -> SyncState:
        return self.build_current_state(context).unwrap()



# stdlib
from collections.abc import Callable
from functools import wraps
import inspect
from typing import Any
from typing import Generic
from typing import ParamSpec
from typing import Set  # noqa: UP035
from typing import cast
from typing import get_args

# third party
from pydantic import ValidationError
import sqlalchemy as sa
from sqlalchemy import Row
from sqlalchemy import Table
from sqlalchemy import func
from sqlalchemy import select
from sqlalchemy.orm import Session
from typing_extensions import Self
from typing_extensions import TypeVar

# relative
from ...serde.json_serde import deserialize_json
from ...serde.json_serde import is_json_primitive
from ...serde.json_serde import serialize_json
from ...server.credentials import SyftVerifyKey
from ...service.action.action_permissions import ActionObjectEXECUTE
from ...service.action.action_permissions import ActionObjectOWNER
from ...service.action.action_permissions import ActionObjectPermission
from ...service.action.action_permissions import ActionObjectREAD
from ...service.action.action_permissions import ActionObjectWRITE
from ...service.action.action_permissions import ActionPermission
from ...service.action.action_permissions import StoragePermission
from ...service.user.user_roles import ServiceRole
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_metaclass import Empty
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SyftObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..document_store_errors import NotFoundException
from ..document_store_errors import StashException
from ..document_store_errors import UniqueConstraintException
from .db import DBManager
from .query import Query
from .schema import PostgresBase
from .schema import SQLiteBase
from .schema import create_table
from .sqlite import SQLiteDBManager

StashT = TypeVar("StashT", bound=SyftObject)
T = TypeVar("T")
P = ParamSpec("P")


def parse_filters(filter_dict: dict[str, Any] | None) -> list[tuple[str, str, Any]]:
    # NOTE using django style filters, e.g. {"age__gt": 18}
    if filter_dict is None:
        return []
    filters = []
    for key, value in filter_dict.items():
        key_split = key.split("__")
        # Operator is eq if not specified
        if len(key_split) == 1:
            field, operator = key, "eq"
        elif len(key_split) == 2:
            field, operator = key_split
        filters.append((field, operator, value))
    return filters


def with_session(func: Callable[P, T]) -> Callable[P, T]:  # type: ignore
    """
    Decorator to inject a session into the function kwargs if it is not provided.

    Make sure to pass session as a keyword argument to the function.

    TODO: This decorator is a temporary fix, we want to move to a DI approach instead:
    move db connection and session to context, and pass context to all stash methods.
    """

    # inspect if the function has a session kwarg
    sig = inspect.signature(func)
    inject_session: bool = "session" in sig.parameters

    @wraps(func)
    def wrapper(self: "ObjectStash[StashT]", *args: Any, **kwargs: Any) -> Any:
        if inject_session and kwargs.get("session") is None:
            with self.sessionmaker() as session:
                with session.begin():
                    kwargs["session"] = session
                    return func(self, *args, **kwargs)
        return func(self, *args, **kwargs)

    return wrapper  # type: ignore


@instrument
class ObjectStash(Generic[StashT]):
    allow_any_type: bool = False

    def __init__(self, store: DBManager) -> None:
        self.db = store
        self.object_type = self.get_object_type()
        self.table = create_table(self.object_type, self.dialect)
        self.sessionmaker: Callable[[], Session] = self.db.sessionmaker

    @property
    def dialect(self) -> sa.engine.interfaces.Dialect:
        return self.db.engine.dialect

    @classmethod
    def get_object_type(cls) -> type[StashT]:
        """
        Get the object type this stash is storing. This is the generic argument of the
        ObjectStash class.
        """
        generic_args = get_args(cls.__orig_bases__[0])
        if len(generic_args) != 1:
            raise TypeError("ObjectStash must have a single generic argument")
        elif not issubclass(generic_args[0], SyftObject):
            raise TypeError(
                "ObjectStash generic argument must be a subclass of SyftObject"
            )
        return generic_args[0]

    @with_session
    def __len__(self, session: Session = None) -> int:
        return session.query(self.table).count()

    @classmethod
    def random(cls, **kwargs: dict) -> Self:
        """Create a random stash with a random server_uid and root_verify_key. Useful for development."""
        db_manager = SQLiteDBManager.random(**kwargs)
        stash = cls(store=db_manager)
        stash.db.init_tables()
        return stash

    def _is_sqlite(self) -> bool:
        return self.db.engine.dialect.name == "sqlite"

    @property
    def server_uid(self) -> UID:
        return self.db.server_uid

    @property
    def root_verify_key(self) -> SyftVerifyKey:
        return self.db.root_verify_key

    @property
    def _data(self) -> list[StashT]:
        return self.get_all(self.root_verify_key, has_permission=True).unwrap()

    def query(self, object_type: type[SyftObject] | None = None) -> Query:
        """Creates a query for this stash's object type and SQL dialect."""
        object_type = object_type or self.object_type
        return Query.create(object_type, self.dialect)

    @as_result(StashException)
    def check_type(self, obj: T, type_: type) -> T:
        if not isinstance(obj, type_):
            raise StashException(f"{type(obj)} does not match required type: {type_}")
        return cast(T, obj)

    @property
    def session(self) -> Session:
        return self.db.session

    def _print_query(self, stmt: sa.sql.select) -> None:
        print(
            stmt.compile(
                compile_kwargs={"literal_binds": True},
                dialect=self.db.engine.dialect,
            )
        )

    @property
    def unique_fields(self) -> list[str]:
        return getattr(self.object_type, "__attr_unique__", [])

    @with_session
    def is_unique(self, obj: StashT, session: Session = None) -> bool:
        unique_fields = self.unique_fields
        if not unique_fields:
            return True

        filters = []
        for field_name in unique_fields:
            field_value = getattr(obj, field_name, None)
            if not is_json_primitive(field_value):
                raise StashException(
                    f"Cannot check uniqueness of non-primitive field {field_name}"
                )
            if field_value is None:
                continue
            filters.append((field_name, "eq", field_value))

        query = self.query()
        query = query.filter_or(
            *filters,
        )

        results = query.execute(session).all()

        if len(results) > 1:
            return False
        elif len(results) == 1:
            result = results[0]
            res = result.id == obj.id
            return res
        return True

    @with_session
    def exists(
        self, credentials: SyftVerifyKey, uid: UID, session: Session = None
    ) -> bool:
        # TODO should be @as_result
        # TODO needs credentials check?
        # TODO use COUNT(*) instead of SELECT
        query = self.query().filter("id", "eq", uid)
        result = query.execute(session).first()
        return result is not None

    @as_result(SyftException, StashException, NotFoundException)
    @with_session
    def get_by_uid(
        self,
        credentials: SyftVerifyKey,
        uid: UID,
        has_permission: bool = False,
        session: Session = None,
    ) -> StashT:
        return self.get_one(
            credentials=credentials,
            filters={"id": uid},
            has_permission=has_permission,
            session=session,
        ).unwrap()

    def _get_field_filter(
        self,
        field_name: str,
        field_value: Any,
        table: Table | None = None,
    ) -> sa.sql.elements.BinaryExpression:
        table = table if table is not None else self.table
        if field_name == "id":
            uid_field_value = UID(field_value)
            return table.c.id == uid_field_value

        json_value = serialize_json(field_value)
        if self.db.engine.dialect.name == "sqlite":
            return table.c.fields[field_name] == func.json_quote(json_value)
        elif self.db.engine.dialect.name == "postgresql":
            return table.c.fields[field_name].astext == cast(json_value, sa.String)

    @as_result(SyftException, StashException, NotFoundException)
    def get_index(
        self, credentials: SyftVerifyKey, index: int, has_permission: bool = False
    ) -> StashT:
        order_by, sort_order = self.object_type.__order_by__
        if index < 0:
            index = -1 - index
            sort_order = "desc" if sort_order == "asc" else "asc"

        items = self.get_all(
            credentials,
            has_permission=has_permission,
            limit=1,
            offset=index,
            order_by=order_by,
            sort_order=sort_order,
        ).unwrap()

        if len(items) == 0:
            raise NotFoundException(f"No item found at index {index}")
        return items[0]

    def row_as_obj(self, row: Row) -> StashT:
        # TODO make unwrappable serde
        return deserialize_json(row.fields)

    @with_session
    def get_role(
        self, credentials: SyftVerifyKey, session: Session = None
    ) -> ServiceRole:
        # relative
        from ...service.user.user import User

        Base = SQLiteBase if self._is_sqlite() else PostgresBase

        # TODO error handling
        if Base.metadata.tables.get("User") is None:
            # if User table does not exist, we assume the user is a guest
            # this happens when we create stashes in tests
            return ServiceRole.GUEST

        try:
            query = self.query(User).filter("verify_key", "eq", credentials)
        except Exception as e:
            print("Error getting role", e)
            raise e

        user = query.execute(session).first()
        if user is None:
            return ServiceRole.GUEST

        return self.row_as_obj(user).role

    def _get_permission_filter_from_permisson(
        self,
        permission: ActionObjectPermission,
    ) -> sa.sql.elements.BinaryExpression:
        permission_string = permission.permission_string
        compound_permission_string = permission.compound_permission_string

        if self.db.engine.dialect.name == "postgresql":
            permission_string = [permission_string]  # type: ignore
            compound_permission_string = [compound_permission_string]  # type: ignore
        return sa.or_(
            self.table.c.permissions.contains(permission_string),
            self.table.c.permissions.contains(compound_permission_string),
        )

    @with_session
    def _apply_permission_filter(
        self,
        stmt: T,
        *,
        credentials: SyftVerifyKey,
        permission: ActionPermission = ActionPermission.READ,
        has_permission: bool = False,
        session: Session = None,
    ) -> T:
        if has_permission:
            # ignoring permissions
            return stmt
        role = self.get_role(credentials, session=session)
        if role in (ServiceRole.ADMIN, ServiceRole.DATA_OWNER):
            # admins and data owners have all permissions
            return stmt

        action_object_permission = ActionObjectPermission(
            uid=UID(),  # dummy uid, we just need the permission string
            credentials=credentials,
            permission=permission,
        )

        stmt = stmt.where(
            self._get_permission_filter_from_permisson(
                permission=action_object_permission
            )
        )
        return stmt

    @as_result(SyftException, StashException)
    @with_session
    def set(
        self,
        credentials: SyftVerifyKey,
        obj: StashT,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,  # TODO: check the default value
        ignore_duplicates: bool = False,
        session: Session = None,
        skip_check_type: bool = False,
    ) -> StashT:
        if not self.allow_any_type and not skip_check_type:
            self.check_type(obj, self.object_type).unwrap()
        uid = obj.id

        # check if the object already exists
        if self.exists(credentials, uid, session=session) or not self.is_unique(
            obj, session=session
        ):
            if ignore_duplicates:
                return obj
            unique_fields_str = ", ".join(self.unique_fields)
            raise UniqueConstraintException(
                public_message=f"Duplication Key Error for {obj}.\n"
                f"The fields that should be unique are {unique_fields_str}."
            )

        permissions = self.get_ownership_permissions(uid, credentials)
        if add_permissions is not None:
            add_permission_strings = [p.permission_string for p in add_permissions]
            permissions.extend(add_permission_strings)

        storage_permissions = []
        if add_storage_permission:
            storage_permissions.append(
                self.server_uid.no_dash,
            )

        fields = serialize_json(obj)
        try:
            # check if the fields are deserializable
            # TODO: Ideally, we want to make sure we don't serialize what we cannot deserialize
            #       and remove this check.
            deserialize_json(fields)
        except Exception as e:
            raise StashException(
                f"Error serializing object: {e}. Some fields are invalid."
            )
        # create the object with the permissions
        stmt = self.table.insert().values(
            id=uid,
            fields=fields,
            permissions=permissions,
            storage_permissions=storage_permissions,
        )
        session.execute(stmt)
        return self.get_by_uid(credentials, uid, session=session).unwrap()

    @as_result(ValidationError, AttributeError)
    def apply_partial_update(
        self, original_obj: StashT, update_obj: SyftObject
    ) -> StashT:
        for key, value in update_obj.__dict__.items():
            if value is Empty:
                continue

            if key in original_obj.__dict__:
                setattr(original_obj, key, value)
            else:
                raise AttributeError(
                    f"{type(update_obj).__name__}.{key} not found in {type(original_obj).__name__}"
                )

        # validate the new fields
        self.object_type.model_validate(original_obj)
        return original_obj

    @as_result(
        StashException,
        NotFoundException,
        AttributeError,
        ValidationError,
        UniqueConstraintException,
    )
    @with_session
    def update(
        self,
        credentials: SyftVerifyKey,
        obj: StashT,
        has_permission: bool = False,
        session: Session = None,
    ) -> StashT:
        """
        NOTE: We cannot do partial updates on the database,
        because we are using computed fields that are not known to the DB:
        - serialize_json will add computed fields to the JSON stored in the database
        - If we update a single field in the JSON, the computed fields can get out of sync.
        - To fix, we either need db-supported computed fields, or know in our ORM which fields should be re-computed.
        """

        if issubclass(type(obj), PartialSyftObject):
            original_obj = self.get_by_uid(
                credentials, obj.id, session=session
            ).unwrap()
            obj = self.apply_partial_update(
                original_obj=original_obj, update_obj=obj
            ).unwrap()

        # TODO has_permission is not used
        if not self.is_unique(obj, session=session):
            raise UniqueConstraintException(
                f"Some fields are not unique for {type(obj).__name__} and unique fields {self.unique_fields}"
            )

        stmt = self.table.update().where(self._get_field_filter("id", obj.id))
        stmt = self._apply_permission_filter(
            stmt,
            credentials=credentials,
            permission=ActionPermission.WRITE,
            has_permission=has_permission,
            session=session,
        )
        fields = serialize_json(obj)
        try:
            deserialize_json(fields)
        except Exception as e:
            raise StashException(
                f"Error serializing object: {e}. Some fields are invalid."
            )
        stmt = stmt.values(fields=fields)
        result = session.execute(stmt)
        if result.rowcount == 0:
            raise NotFoundException(
                f"{self.object_type.__name__}: {obj.id} not found or no permission to update."
            )
        return self.get_by_uid(credentials, obj.id, session=session).unwrap()

    @as_result(StashException, NotFoundException)
    @with_session
    def delete_by_uid(
        self,
        credentials: SyftVerifyKey,
        uid: UID,
        has_permission: bool = False,
        session: Session = None,
    ) -> UID:
        stmt = self.table.delete().where(self._get_field_filter("id", uid))
        stmt = self._apply_permission_filter(
            stmt,
            credentials=credentials,
            permission=ActionPermission.WRITE,
            has_permission=has_permission,
            session=session,
        )
        result = session.execute(stmt)
        if result.rowcount == 0:
            raise NotFoundException(
                f"{self.object_type.__name__}: {uid} not found or no permission to delete."
            )
        return uid

    @as_result(StashException)
    @with_session
    def get_one(
        self,
        credentials: SyftVerifyKey,
        filters: dict[str, Any] | None = None,
        has_permission: bool = False,
        order_by: str | None = None,
        sort_order: str | None = None,
        offset: int = 0,
        session: Session = None,
    ) -> StashT:
        """
        Get first objects from the stash, optionally filtered.

        Args:
            credentials (SyftVerifyKey): credentials of the user
            filters (dict[str, Any] | None, optional): dictionary of filters,
                where the key is the field name and the value is the filter value.
                Operators other than equals can be used in the key,
                e.g. {"name": "Bob", "friends__contains": "Alice"}. Defaults to None.
            has_permission (bool, optional): If True, overrides the permission check.
                Defaults to False.
            order_by (str | None, optional): If provided, the results will be ordered by this field.
                If not provided, the default order and field defined on the SyftObject.__order_by__ are used.
                Defaults to None.
            sort_order (str | None, optional): "asc" or "desc" If not defined,
                the default order defined on the SyftObject.__order_by__ is used.
                Defaults to None.
            offset (int, optional): offset the results. Defaults to 0.

        Returns:
            list[StashT]: list of objects.
        """
        query = self.query()

        if not has_permission:
            role = self.get_role(credentials, session=session)
            query = query.with_permissions(credentials, role)

        for field_name, operator, field_value in parse_filters(filters):
            query = query.filter(field_name, operator, field_value)

        query = query.order_by(order_by, sort_order).offset(offset).limit(1)
        result = query.execute(session).first()
        if result is None:
            raise NotFoundException(f"{self.object_type.__name__}: not found")

        return self.row_as_obj(result)

    @as_result(StashException)
    @with_session
    def get_all(
        self,
        credentials: SyftVerifyKey,
        filters: dict[str, Any] | None = None,
        has_permission: bool = False,
        order_by: str | None = None,
        sort_order: str | None = None,
        limit: int | None = None,
        offset: int = 0,
        session: Session = None,
    ) -> list[StashT]:
        """
        Get all objects from the stash, optionally filtered.

        Args:
            credentials (SyftVerifyKey): credentials of the user
            filters (dict[str, Any] | None, optional): dictionary of filters,
                where the key is the field name and the value is the filter value.
                Operators other than equals can be used in the key,
                e.g. {"name": "Bob", "friends__contains": "Alice"}. Defaults to None.
            has_permission (bool, optional): If True, overrides the permission check.
                Defaults to False.
            order_by (str | None, optional): If provided, the results will be ordered by this field.
                If not provided, the default order and field defined on the SyftObject.__order_by__ are used.
                Defaults to None.
            sort_order (str | None, optional): "asc" or "desc" If not defined,
                the default order defined on the SyftObject.__order_by__ is used.
                Defaults to None.
            limit (int | None, optional): limit the number of results. Defaults to None.
            offset (int, optional): offset the results. Defaults to 0.

        Returns:
            list[StashT]: list of objects.
        """
        query = self.query()

        if not has_permission:
            role = self.get_role(credentials, session=session)
            query = query.with_permissions(credentials, role)

        for field_name, operator, field_value in parse_filters(filters):
            query = query.filter(field_name, operator, field_value)

        query = query.order_by(order_by, sort_order).limit(limit).offset(offset)
        result = query.execute(session).all()
        return [self.row_as_obj(row) for row in result]

    # PERMISSIONS
    def get_ownership_permissions(
        self, uid: UID, credentials: SyftVerifyKey
    ) -> list[str]:
        return [
            ActionObjectOWNER(uid=uid, credentials=credentials).permission_string,
            ActionObjectWRITE(uid=uid, credentials=credentials).permission_string,
            ActionObjectREAD(uid=uid, credentials=credentials).permission_string,
            ActionObjectEXECUTE(uid=uid, credentials=credentials).permission_string,
        ]

    @as_result(NotFoundException)
    @with_session
    def add_permission(
        self,
        permission: ActionObjectPermission,
        session: Session = None,
        ignore_missing: bool = False,
    ) -> None:
        try:
            existing_permissions = self._get_permissions_for_uid(
                permission.uid, session=session
            ).unwrap()
        except NotFoundException:
            if ignore_missing:
                return None
            raise

        existing_permissions.add(permission.permission_string)

        stmt = self.table.update().where(self.table.c.id == permission.uid)
        stmt = stmt.values(permissions=list(existing_permissions))
        session.execute(stmt)
        return None

    @as_result(NotFoundException)
    @with_session
    def add_permissions(
        self,
        permissions: list[ActionObjectPermission],
        ignore_missing: bool = False,
        session: Session = None,
    ) -> None:
        for permission in permissions:
            self.add_permission(
                permission, session=session, ignore_missing=ignore_missing
            ).unwrap()
        return None

    @with_session
    def remove_permission(
        self, permission: ActionObjectPermission, session: Session = None
    ) -> None:
        # TODO not threadsafe
        try:
            permissions = self._get_permissions_for_uid(permission.uid).unwrap()
            permissions.remove(permission.permission_string)
        except (NotFoundException, KeyError):
            # TODO add error handling to permissions
            return None

        stmt = (
            self.table.update()
            .where(self.table.c.id == permission.uid)
            .values(permissions=list(permissions))
        )
        session.execute(stmt)
        return None

    @with_session
    def has_permission(
        self, permission: ActionObjectPermission, session: Session = None
    ) -> bool:
        if self.get_role(permission.credentials, session=session) in (
            ServiceRole.ADMIN,
            ServiceRole.DATA_OWNER,
        ):
            return True
        return self.has_permissions([permission], session=session)

    @with_session
    def has_permissions(
        self, permissions: list[ActionObjectPermission], session: Session = None
    ) -> bool:
        # TODO: we should use a permissions table to check all permissions at once

        permission_filters = [
            sa.and_(
                self._get_field_filter("id", p.uid),
                self._get_permission_filter_from_permisson(permission=p),
            )
            for p in permissions
        ]

        stmt = self.table.select().where(
            sa.and_(
                *permission_filters,
            ),
        )
        result = session.execute(stmt).first()
        return result is not None

    @as_result(StashException)
    @with_session
    def _get_permissions_for_uid(self, uid: UID, session: Session = None) -> Set[str]:  # noqa: UP006
        stmt = select(self.table.c.permissions).where(self.table.c.id == uid)
        result = session.execute(stmt).scalar_one_or_none()
        if result is None:
            raise NotFoundException(f"No permissions found for uid: {uid}")
        return set(result)

    @as_result(StashException)
    @with_session
    def get_all_permissions(self, session: Session = None) -> dict[UID, Set[str]]:  # noqa: UP006
        stmt = select(self.table.c.id, self.table.c.permissions)
        results = session.execute(stmt).all()
        return {UID(row.id): set(row.permissions) for row in results}

    # STORAGE PERMISSIONS
    @with_session
    def has_storage_permission(
        self, permission: StoragePermission, session: Session = None
    ) -> bool:
        return self.has_storage_permissions([permission], session=session)

    @with_session
    def has_storage_permissions(
        self, permissions: list[StoragePermission], session: Session = None
    ) -> bool:
        permission_filters = [
            sa.and_(
                self._get_field_filter("id", p.uid),
                self.table.c.storage_permissions.contains(
                    p.server_uid.no_dash
                    if self._is_sqlite()
                    else [p.server_uid.no_dash]
                ),
            )
            for p in permissions
        ]

        stmt = self.table.select().where(
            sa.and_(
                *permission_filters,
            )
        )
        result = session.execute(stmt).first()
        return result is not None

    @as_result(StashException)
    @with_session
    def get_all_storage_permissions(
        self, session: Session = None
    ) -> dict[UID, Set[UID]]:  # noqa: UP006
        stmt = select(self.table.c.id, self.table.c.storage_permissions)
        results = session.execute(stmt).all()

        return {
            UID(row.id): {UID(uid) for uid in row.storage_permissions}
            for row in results
        }

    @as_result(NotFoundException)
    @with_session
    def add_storage_permissions(
        self,
        permissions: list[StoragePermission],
        session: Session = None,
        ignore_missing: bool = False,
    ) -> None:
        for permission in permissions:
            self.add_storage_permission(
                permission, session=session, ignore_missing=ignore_missing
            ).unwrap()

        return None

    @as_result(NotFoundException)
    @with_session
    def add_storage_permission(
        self,
        permission: StoragePermission,
        session: Session = None,
        ignore_missing: bool = False,
    ) -> None:
        try:
            existing_permissions = self._get_storage_permissions_for_uid(
                permission.uid, session=session
            ).unwrap()
        except NotFoundException:
            if ignore_missing:
                return None
            raise

        existing_permissions.add(permission.server_uid)

        stmt = (
            self.table.update()
            .where(self.table.c.id == permission.uid)
            .values(storage_permissions=[str(uid) for uid in existing_permissions])
        )

        session.execute(stmt)

    @with_session
    def remove_storage_permission(
        self, permission: StoragePermission, session: Session = None
    ) -> None:
        try:
            permissions = self._get_storage_permissions_for_uid(
                permission.uid, session=session
            ).unwrap()
            permissions.discard(permission.server_uid)
        except NotFoundException:
            # TODO add error handling to permissions
            return None

        stmt = (
            self.table.update()
            .where(self.table.c.id == permission.uid)
            .values(storage_permissions=[str(uid) for uid in permissions])
        )
        session.execute(stmt)
        return None

    @as_result(StashException)
    @with_session
    def _get_storage_permissions_for_uid(
        self, uid: UID, session: Session = None
    ) -> Set[UID]:  # noqa: UP006
        stmt = select(self.table.c.id, self.table.c.storage_permissions).where(
            self.table.c.id == uid
        )
        result = session.execute(stmt).first()
        if result is None:
            raise NotFoundException(f"No storage permissions found for uid: {uid}")
        return {UID(uid) for uid in result.storage_permissions}

    @with_session
    @as_result(StashException)
    def upsert(
        self,
        credentials: SyftVerifyKey,
        obj: StashT,
        session: Session = None,
    ) -> StashT:
        """Insert or update an object in the stash if it already exists.
        Atomic operation when using the same session for both operations.
        """

        try:
            return self.set(
                credentials=credentials,
                obj=obj,
                session=session,
            ).unwrap()
        except UniqueConstraintException:
            return self.update(
                credentials=credentials, obj=obj, session=session
            ).unwrap()



def is_within_rate_limit(context) -> bool:
    """Rate limiter for custom API calls made by users."""
    # stdlib
    import datetime

    state = context.state
    settings = context.settings
    email = context.user.email

    current_time = datetime.datetime.now()
    calls_last_min = [
        1 if (current_time - call_time).seconds < 60 else 0
        for call_time in state[email]
    ]

    return sum(calls_last_min) < settings.get("calls_per_min", 5)



# stdlib
from math import nan

schema_dict = {
    "project": {
        0: "example-project",
        1: "example-project",
        2: "example-project",
        3: "example-project",
        4: "example-project",
        5: "example-project",
        6: "example-project",
        7: "example-project",
        8: "example-project",
        9: "example-project",
        10: "example-project",
        11: "example-project",
        12: "example-project",
        13: "example-project",
        14: "example-project",
        15: "example-project",
        16: "example-project",
        17: "example-project",
        18: "example-project",
        19: "example-project",
        20: "example-project",
        21: "example-project",
        22: "example-project",
    },
    "dataset_id": {
        0: "test_1gb",
        1: "test_1gb",
        2: "test_1gb",
        3: "test_1gb",
        4: "test_1gb",
        5: "test_1gb",
        6: "test_1gb",
        7: "test_1gb",
        8: "test_1gb",
        9: "test_1gb",
        10: "test_1gb",
        11: "test_1gb",
        12: "test_1gb",
        13: "test_1gb",
        14: "test_1gb",
        15: "test_1gb",
        16: "test_1gb",
        17: "test_1gb",
        18: "test_1gb",
        19: "test_1gb",
        20: "test_1gb",
        21: "test_1gb",
        22: "test_1gb",
    },
    "table_id": {
        0: "posts",
        1: "posts",
        2: "posts",
        3: "posts",
        4: "posts",
        5: "posts",
        6: "posts",
        7: "comments",
        8: "comments",
        9: "comments",
        10: "comments",
        11: "comments",
        12: "comments",
        13: "comments",
        14: "comments",
        15: "comments",
        16: "comments",
        17: "comments",
        18: "comments",
        19: "comments",
        20: "comments",
        21: "comments",
        22: "comments",
    },
    "schema_name": {
        0: "int64_field_0",
        1: "id",
        2: "name",
        3: "subscribers_count",
        4: "permalink",
        5: "nsfw",
        6: "spam",
        7: "int64_field_0",
        8: "id",
        9: "body",
        10: "parent_id",
        11: "created_at",
        12: "last_modified_at",
        13: "gilded",
        14: "permalink",
        15: "score",
        16: "comment_id",
        17: "post_id",
        18: "author_id",
        19: "spam",
        20: "deleted",
        21: "upvote_raio",
        22: "collapsed_in_crowd_control",
    },
    "schema_field": {
        0: "INTEGER",
        1: "STRING",
        2: "STRING",
        3: "INTEGER",
        4: "STRING",
        5: "FLOAT",
        6: "BOOLEAN",
        7: "INTEGER",
        8: "STRING",
        9: "STRING",
        10: "STRING",
        11: "INTEGER",
        12: "INTEGER",
        13: "BOOLEAN",
        14: "STRING",
        15: "INTEGER",
        16: "STRING",
        17: "STRING",
        18: "STRING",
        19: "BOOLEAN",
        20: "BOOLEAN",
        21: "FLOAT",
        22: "BOOLEAN",
    },
    "description": {
        0: "None",
        1: "None",
        2: "None",
        3: "None",
        4: "None",
        5: "None",
        6: "None",
        7: "None",
        8: "None",
        9: "None",
        10: "None",
        11: "None",
        12: "None",
        13: "None",
        14: "None",
        15: "None",
        16: "None",
        17: "None",
        18: "None",
        19: "None",
        20: "None",
        21: "None",
        22: "None",
    },
    "num_rows": {
        0: "2000000",
        1: "2000000",
        2: "2000000",
        3: "2000000",
        4: "2000000",
        5: "2000000",
        6: "2000000",
        7: "2000000",
        8: "2000000",
        9: "2000000",
        10: "2000000",
        11: "2000000",
        12: "2000000",
        13: "2000000",
        14: "2000000",
        15: "2000000",
        16: "2000000",
        17: "2000000",
        18: "2000000",
        19: "2000000",
        20: "2000000",
        21: "2000000",
        22: "2000000",
    },
}


query_dict = {
    "int64_field_0": {
        0: 4,
        1: 5,
        2: 10,
        3: 16,
        4: 17,
        5: 23,
        6: 24,
        7: 25,
        8: 27,
        9: 40,
    },
    "id": {
        0: "t5_via1x",
        1: "t5_cv9gn",
        2: "t5_8p2tq",
        3: "t5_8fcro",
        4: "t5_td5of",
        5: "t5_z01fv",
        6: "t5_hmqjk",
        7: "t5_1flyj",
        8: "t5_5rwej",
        9: "t5_uurcv",
    },
    "name": {
        0: "/channel/mylittlepony",
        1: "/channel/polyamory",
        2: "/channel/Catholicism",
        3: "/channel/cordcutters",
        4: "/channel/stevenuniverse",
        5: "/channel/entitledbitch",
        6: "/channel/engineering",
        7: "/channel/nottheonion",
        8: "/channel/FoodPorn",
        9: "/channel/puppysmiles",
    },
    "subscribers_count": {
        0: 4323081,
        1: 2425929,
        2: 4062607,
        3: 7543226,
        4: 2692168,
        5: 2709080,
        6: 8766144,
        7: 2580984,
        8: 7784809,
        9: 3715991,
    },
    "permalink": {
        0: "/channel//channel/mylittlepony",
        1: "/channel//channel/polyamory",
        2: "/channel//channel/Catholicism",
        3: "/channel//channel/cordcutters",
        4: "/channel//channel/stevenuniverse",
        5: "/channel//channel/entitledbitch",
        6: "/channel//channel/engineering",
        7: "/channel//channel/nottheonion",
        8: "/channel//channel/FoodPorn",
        9: "/channel//channel/puppysmiles",
    },
    "nsfw": {
        0: nan,
        1: nan,
        2: nan,
        3: nan,
        4: nan,
        5: nan,
        6: nan,
        7: nan,
        8: nan,
        9: nan,
    },
    "spam": {
        0: False,
        1: False,
        2: False,
        3: False,
        4: False,
        5: False,
        6: False,
        7: False,
        8: False,
        9: False,
    },
}



# stdlib
from collections.abc import Callable
from importlib import import_module
import inspect
import os
import traceback
from types import CodeType
from types import TracebackType
from typing import Any
from typing import TYPE_CHECKING
from typing import TypeVar
import uuid

# third party
from IPython import get_ipython
from IPython.display import HTML
from IPython.display import display
import psutil
from typing_extensions import Self

# relative
from ..service.user.user_roles import ServiceRole
from ..util.notebook_ui.components.tabulator_template import jinja_env

if TYPE_CHECKING:
    # relative
    from ..service.context import AuthedServiceContext


class SyftException(Exception):
    """
    A Syft custom exception class with distinct public and private messages.

    Attributes:
        private_message (str): Detailed error message intended for administrators.
        public_message (str): General error message for end-users.
    """

    public_message = "An error occurred. Contact the admin for more information."

    def __init__(
        self,
        private_message: str | None = None,
        public_message: str | None = None,
        server_trace: str | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if public_message is not None and not isinstance(public_message, str):
            raise TypeError("public message should be a string")
        if private_message is not None and not isinstance(private_message, str):
            raise TypeError("private message should be a string")

        if public_message:
            self.public_message = public_message

        self._private_message = private_message or ""
        self._server_trace = server_trace or ""
        super().__init__(self.public, *args, **kwargs)

    @property
    def public(self) -> str:
        """
        Returns the public error message.

        Returns:
            str: The public error message.
        """
        return self.public_message

    def get_message(self, context: "AuthedServiceContext") -> str:
        """
        Retrieves the appropriate message based on the user's role, obtained via
        `context.role`.

        Args:
            context (AuthedServiceContext): The server context.

        Returns:
            str: The private or public message based on the role.
        """
        if context.role.value >= ServiceRole.DATA_OWNER.value or context.dev_mode:
            return self._private_message or self.public
        return self.public

    def get_tb(
        self,
        context: "AuthedServiceContext | None" = None,
        overwrite_permission: bool = False,
    ) -> str | None:
        """
        Returns the error traceback as a string, if the user is able to see it.

        Args:
            context (AuthedServiceContext): The authenticated service context which
                contains the user's role.

        Returns:
            str | None: A string representation of the current stack trace if the
                user is a DataOwner or higher, otherwise None.
        """
        # stdlib
        import traceback

        if (
            overwrite_permission
            or (context and context.role.value >= ServiceRole.DATA_OWNER.value)
            or (context and context.dev_mode)
        ):
            return "".join(traceback.format_exception(self))
        return None

    @classmethod
    def from_exception(
        cls,
        exc: BaseException,
        public_message: str | None = None,
        private_message: str | None = None,
    ) -> Self:
        """
        Builds an instance of SyftException from an existing exception. It allows
        setting a public message for end users or resetting the private message.
        If no private_message is provided, the original exception's message is used.
        If no public_message is provided, the default public message is used.

        Args:
            exception (BaseException): The original exception from which to create
                the new instance. The message from this exception will be used as
                the base message for the new instance.

            public_message (str, optional): An optional message intended for public
                display. This message can provide user-friendly information about
                the error. If not provided, the default message is used.

            private_message (str, optional): An optional message intended for private
                display. This message should provide more information about the error
                to administrators. If not provided, the exception's message is used.

        Returns:
            Self: A new instance of the class. The new instance retains the traceback
                of the original exception.
        """
        if isinstance(exc, SyftException):
            private_message = private_message or exc._private_message
            public_message = public_message or exc.public_message
        elif isinstance(exc, BaseException):
            private_message = private_message or str(exc)

        new_exc = cls(private_message, public_message=public_message)
        new_exc.__traceback__ = exc.__traceback__
        new_exc.__cause__ = exc
        new_exc = process_traceback(new_exc)
        return new_exc

    @property
    def _repr_html_class_(self) -> str:
        return "alert-danger"

    def __str__(self) -> str:
        # this assumes that we show the server side error on the client side without a jupyter notebook
        server_trace = self._server_trace
        message = self._private_message or self.public

        if server_trace:
            message = f"{message}\nserver_trace: {server_trace}"

        return message

    def _repr_html_(self) -> str:
        is_dev_mode = os.getenv("DEV_MODE", "false").lower() == "true"
        display = "block" if self._server_trace or is_dev_mode else "none"

        exc = process_traceback(self)
        _traceback_str_list = traceback.format_exception(exc)
        traceback_str = "".join(_traceback_str_list)

        table_template = jinja_env.get_template("syft_exception.jinja2")
        table_html = table_template.render(
            name=type(self).__name__,
            html_id=uuid.uuid4().hex,
            server_trace=self._server_trace,
            message=self._private_message or self.public,
            traceback_str=traceback_str,
            display=display,
            dev_mode=is_dev_mode,
        )
        return table_html


class raises:
    def __init__(self, expected_exception, show=False):  # type: ignore
        self.expected_exception = expected_exception
        self.show = show

    def __enter__(self):  # type: ignore
        # Before block of code
        pass

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore
        message = None
        expected_exception_type = self.expected_exception
        if not isinstance(expected_exception_type, type):
            expected_exception_type = type(self.expected_exception)
            if hasattr(self.expected_exception, "public_message"):
                message = self.expected_exception.public_message.replace("*", "")

        # After block of code
        if exc_type is None:
            raise AssertionError(
                f"Expected {self.expected_exception} to be raised, "
                "but no exception was raised."
            )
        if not issubclass(exc_type, expected_exception_type):
            raise AssertionError(
                f"Expected {expected_exception_type} to be raised, but got {exc_type}."
            )
        if message and message not in exc_value.public_message:
            raise AssertionError(
                f"Expected {expected_exception_type} to be raised, "
                f"did not contain {message}."
            )
        if self.show:
            # keep this print!
            print("with sy.raises successfully caught the following exception:")
            if hasattr(exc_value, "_repr_html_"):
                display(HTML(exc_value._repr_html_()))
            else:
                print(
                    f"The following exception was catched\n{exc_value}",
                )
        return True  # Suppress the exception


class ExceptionFilter(tuple):
    """
    Filter and store all exception classes from a given module path. This class can be
    used in try/except blocks to handle these exceptions as a group (see example).

    Attributes:
        module (str): The name of the module from which exceptions are filtered.

    Example:
        ```
        from syft.types.errors import ExceptionFilter

        try:
            ...
        except ExceptionFilter("google.cloud.bigquery") as e:
            ...
        ```
    """

    def __init__(self, module: str) -> None:
        self.module = module

    def __new__(cls, module: str) -> Self:
        """
        Creates a new instance of ExceptionFilter, which gathers all exception classes
        from the specified module and stores them as a tuple.
        """
        exceptions: tuple[type[BaseException], ...]

        try:
            imported_module = import_module(module)
        except ModuleNotFoundError:
            # TODO: log warning
            exceptions = ()
        else:
            exceptions = tuple(
                obj
                for _, obj in inspect.getmembers(imported_module, inspect.isclass)
                if issubclass(obj, BaseException) and not issubclass(obj, Warning)
            )

        instance = super().__new__(cls, exceptions)

        instance.module = module

        return instance


_excluded_code_objects: set[CodeType] = set()

E = TypeVar("E", bound=BaseException)
F = TypeVar("F", bound=Callable[..., object])


def exclude_from_traceback(f: F) -> F:
    """
    Decorator to mark a function to be removed from the traceback when an
    exception is raised. This is useful for functions that are not relevant to
    the error message and would only clutter the traceback.
    """
    _excluded_code_objects.add(f.__code__)
    return f


def process_traceback(exc: E) -> E:
    """
    Adjusts the traceback of an exception to remove specific frames related to
    the as_result decorator and the unwrap() call, for cleaner and more relevant
    error messages.

    Args:
        exc (E): The exception whose traceback is to be adjusted.

    Returns:
        E: The same exception with an adjusted traceback.
    """
    # We want to adjust the traceback so we can remove frames which contain
    # a function marked with the _excluded_code_objects decorator, improving
    # the stacktrace of the error messages.
    tb = exc.__traceback__
    frames: list[TracebackType] = []

    while tb is not None:
        if tb.tb_frame.f_code not in _excluded_code_objects:
            frames.append(tb)
        tb = tb.tb_next

    # Before being done, we need to adjust the traceback.tb_next so that
    # the frames are linked together properly.
    for i, tb in enumerate(frames):
        if i + 1 == len(frames):
            tb.tb_next = None
        else:
            tb.tb_next = frames[i + 1]

    return exc


class CredentialsError(SyftException):
    public_message = "Invalid credentials."


def syft_exception_handler(
    shell: Any, etype: Any, evalue: Any, tb: Any, tb_offset: Any = None
) -> None:
    display(HTML(evalue._repr_html_()))


runs_in_pytest = False
for pid in psutil.pids():
    try:
        if "PYTEST_CURRENT_TEST" in psutil.Process(pid).environ():
            runs_in_pytest = True
    except Exception:
        pass  # nosec


# be very careful when changing this. pytest (with nbmake) will
# not pick up exceptions if they have a custom exception handler (fail silently)
if not runs_in_pytest:
    try:
        get_ipython().set_custom_exc((SyftException,), syft_exception_handler)  # noqa: F821
    except Exception:
        pass  # nosec



# relative
from ...serde.serializable import serializable
from ...store.db.stash import ObjectStash
from ...util.telemetry import instrument
from .settings import ServerSettings


@instrument
@serializable(canonical_name="SettingsStashSQL", version=1)
class SettingsStash(ObjectStash[ServerSettings]):
    pass



# stdlib
from typing import Literal

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from .image_registry import SyftImageRegistry


@serializable(canonical_name="SyftImageRegistrySQLStash", version=1)
class SyftImageRegistryStash(ObjectStash[SyftImageRegistry]):
    @as_result(SyftException, StashException, NotFoundException)
    def get_by_url(
        self,
        credentials: SyftVerifyKey,
        url: str,
    ) -> SyftImageRegistry:
        return self.get_one(
            credentials=credentials,
            filters={"url": url},
        ).unwrap()

    @as_result(SyftException, StashException)
    def delete_by_url(self, credentials: SyftVerifyKey, url: str) -> Literal[True]:
        item = self.get_by_url(credentials=credentials, url=url).unwrap()
        self.delete_by_uid(credentials=credentials, uid=item.id).unwrap()

        # TODO standardize delete return type
        return True



# stdlib
import logging
from typing import Any

# third party
import pydantic

# relative
from ...custom_worker.config import DockerWorkerConfig
from ...custom_worker.config import PrebuiltWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...custom_worker.k8s import IN_KUBERNETES
from ...custom_worker.runner_k8s import KubernetesRunner
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...store.linked_obj import LinkedObject
from ...types.dicttuple import DictTuple
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..request.request import Change
from ..request.request import CreateCustomImageChange
from ..request.request import CreateCustomWorkerPoolChange
from ..request.request import Request
from ..request.request import SubmitRequest
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .image_identifier import SyftWorkerImageIdentifier
from .utils import DEFAULT_WORKER_POOL_NAME
from .utils import get_orchestration_type
from .utils import run_containers
from .utils import run_workers_in_threads
from .utils import scale_kubernetes_pool
from .worker_image import SyftWorkerImage
from .worker_image_stash import SyftWorkerImageStash
from .worker_pool import ContainerSpawnStatus
from .worker_pool import WorkerPool
from .worker_pool_stash import SyftWorkerPoolStash
from .worker_service import WorkerService
from .worker_stash import WorkerStash

logger = logging.getLogger(__name__)


@serializable(canonical_name="SyftWorkerPoolService", version=1)
class SyftWorkerPoolService(AbstractService):
    stash: SyftWorkerPoolStash

    def __init__(self, store: DBManager) -> None:
        self.stash = SyftWorkerPoolStash(store=store)
        self.image_stash = SyftWorkerImageStash(store=store)

    @as_result(StashException)
    def pool_exists(self, context: AuthedServiceContext, pool_name: str) -> bool:
        try:
            self.stash.get_by_name(context.credentials, pool_name=pool_name).unwrap()
            return True
        except NotFoundException:
            return False

    @as_result(StashException)
    def image_exists(self, context: AuthedServiceContext, uid: UID) -> bool:
        try:
            self.image_stash.get_by_uid(context.credentials, uid=uid).unwrap()
            return True
        except NotFoundException:
            return False

    @service_method(
        path="worker_pool.launch",
        name="launch",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def launch(
        self,
        context: AuthedServiceContext,
        pool_name: str,
        image_uid: UID | None,
        num_workers: int,
        registry_username: str | None = None,
        registry_password: str | None = None,
        pod_annotations: dict[str, str] | None = None,
        pod_labels: dict[str, str] | None = None,
    ) -> list[ContainerSpawnStatus]:
        """Creates a pool of workers from the given SyftWorkerImage.

        - Retrieves the image for the given UID
        - Use docker to launch containers for given image
        - For each successful container instantiation create a SyftWorker object
        - Creates a SyftWorkerPool object

        Args:
            context (AuthedServiceContext): context passed to the service
            name (str): name of the pool
            image_id (UID): UID of the SyftWorkerImage against which the pool should be created
            num_workers (int): the number of SyftWorker that needs to be created in the pool
        """

        pool_exists = self.pool_exists(context, pool_name=pool_name).unwrap()
        if pool_exists:
            raise SyftException(
                public_message=f"Worker Pool with name: {pool_name} already exists !!"
            )

        # If image uid is not passed, then use the default worker image
        # to create the worker pool
        if image_uid is None:
            default_worker_pool = self.stash.get_by_name(
                context.credentials, pool_name=DEFAULT_WORKER_POOL_NAME
            ).unwrap()
            image_uid = default_worker_pool.image_id

        # Get the image object for the given image id
        worker_image = self.image_stash.get_by_uid(
            credentials=context.credentials, uid=image_uid
        ).unwrap()

        worker_stash = context.server.services.worker.stash

        # Create worker pool from given image, with the given worker pool
        # and with the desired number of workers
        worker_list, container_statuses = _create_workers_in_pool(
            context=context,
            pool_name=pool_name,
            existing_worker_cnt=0,
            worker_cnt=num_workers,
            worker_image=worker_image,
            worker_stash=worker_stash,
            registry_username=registry_username,
            registry_password=registry_password,
            pod_annotations=pod_annotations,
            pod_labels=pod_labels,
        ).unwrap()

        # Update the Database with the pool information
        worker_pool = WorkerPool(
            name=pool_name,
            max_count=num_workers,
            image_id=worker_image.id,
            worker_list=worker_list,
            syft_server_location=context.server.id,
            syft_client_verify_key=context.credentials,
        )
        self.stash.set(credentials=context.credentials, obj=worker_pool).unwrap()
        return container_statuses

    @service_method(
        path="worker_pool.create_pool_request",
        name="pool_creation_request",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def create_pool_request(
        self,
        context: AuthedServiceContext,
        pool_name: str,
        num_workers: int,
        image_uid: UID,
        reason: str | None = "",
        pod_annotations: dict[str, str] | None = None,
        pod_labels: dict[str, str] | None = None,
    ) -> Request:
        """
        Create a request to launch the worker pool based on a built image.

        Args:
            context (AuthedServiceContext): The authenticated service context.
            pool_name (str): The name of the worker pool.
            num_workers (int): The number of workers in the pool.
            image_uid (Optional[UID]): The UID of the built image.
            reason (Optional[str], optional): The reason for creating the
                worker pool. Defaults to "".
        """
        # Check if image exists for the given image id
        worker_image_exists = self.image_exists(context, uid=image_uid).unwrap()

        # Raise error if worker image doesn't exists
        if not worker_image_exists:
            raise SyftException(
                public_message=f"No image exists for given image uid : {image_uid}"
            )

        # Check if pool already exists for the given pool name
        worker_pool_exists = self.pool_exists(context, pool_name=pool_name).unwrap()

        if worker_pool_exists:
            raise SyftException(
                public_message=f"Worker pool already exists for given pool name: {pool_name}"
            )

        # If no worker pool exists for given pool name
        # and image exists for given image uid, then create a change
        # request object to create the pool with the desired number of workers
        create_worker_pool_change = CreateCustomWorkerPoolChange(
            pool_name=pool_name,
            num_workers=num_workers,
            image_uid=image_uid,
            pod_annotations=pod_annotations,
            pod_labels=pod_labels,
        )
        changes: list[Change] = [create_worker_pool_change]

        # Create a the request object with the changes and submit it
        # for approval.
        request = SubmitRequest(changes=changes)
        return context.server.services.request.submit(
            context=context, request=request, reason=reason
        )

    @service_method(
        path="worker_pool.create_image_and_pool_request",
        name="create_image_and_pool_request",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def create_image_and_pool_request(
        self,
        context: AuthedServiceContext,
        pool_name: str,
        num_workers: int,
        config: WorkerConfig,
        tag: str | None = None,
        registry_uid: UID | None = None,
        reason: str | None = "",
        pull_image: bool = True,
        pod_annotations: dict[str, str] | None = None,
        pod_labels: dict[str, str] | None = None,
    ) -> Request:
        """
        Create a request to launch the worker pool based on a built image.

        Args:
            context (AuthedServiceContext): The authenticated service context.
            pool_name (str): The name of the worker pool.
            num_workers (int): The number of workers in the pool.
            config: (WorkerConfig): Config of the image to be built.
            tag (str | None, optional):
                a human-readable manifest identifier that is typically a specific version or variant of an image,
                only needed for `DockerWorkerConfig` to tag the image after it is built.
            reason (str | None, optional): The reason for creating the worker image and pool. Defaults to "".
        """
        if not isinstance(config, DockerWorkerConfig | PrebuiltWorkerConfig):
            raise SyftException(
                public_message="We only support either `DockerWorkerConfig` or `PrebuiltWorkerConfig`."
            )

        if isinstance(config, DockerWorkerConfig):
            if tag is None:
                raise SyftException(
                    public_message="`tag` is required for `DockerWorkerConfig`."
                )

            # Validate image tag
            try:
                SyftWorkerImageIdentifier.from_str(tag=tag)
            except pydantic.ValidationError as e:
                raise SyftException(public_message=f"Invalid `tag`: {e}.")

            if IN_KUBERNETES and registry_uid is None:
                raise SyftException(
                    public_message="`registry_uid` is required in Kubernetes mode for `DockerWorkerConfig`."
                )

        # Check if an image already exists for given docker config
        worker_image_exists = self.image_stash.worker_config_exists(
            credentials=context.credentials, config=config
        ).unwrap()

        if worker_image_exists:
            raise SyftException(
                public_message="Image already exists for given config. \
                    Please use `worker_pool.create_pool_request` to request pool creation."
            )

        # create a list of Change objects and submit a
        # request for these changes for approval
        changes: list[Change] = []

        # Add create custom image change
        # If this change is approved, then build an image using the config
        create_custom_image_change = CreateCustomImageChange(
            config=config,
            tag=tag,
            registry_uid=registry_uid,
            pull_image=pull_image,
        )

        # Check if a pool already exists for given pool name
        worker_pool_exists = self.pool_exists(context, pool_name=pool_name).unwrap()

        # Raise an error if worker pool already exists for the given worker pool name
        if worker_pool_exists:
            raise SyftException(
                public_message=f"Worker Pool with name: {pool_name} already"
                f" exists. Please choose another name!"
            )

        # Add create worker pool change
        # If change is approved then worker pool is created and
        # the desired number of workers are added to the pool
        create_worker_pool_change = CreateCustomWorkerPoolChange(
            pool_name=pool_name,
            num_workers=num_workers,
            config=config,
            pod_annotations=pod_annotations,
            pod_labels=pod_labels,
        )
        changes += [create_custom_image_change, create_worker_pool_change]

        # Create a request object and submit a request for approval
        request = SubmitRequest(changes=changes)
        return context.server.services.request.submit(
            context=context, request=request, reason=reason
        )

    @service_method(
        path="worker_pool.get_all",
        name="get_all",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all(self, context: AuthedServiceContext) -> DictTuple[str, WorkerPool]:
        # TODO: During get_all, we should dynamically make a call to docker to get the status of the containers
        # and update the status of the workers in the pool.
        worker_pools = self.stash.get_all(credentials=context.credentials).unwrap()

        res = ((pool.name, pool) for pool in worker_pools)
        return DictTuple(res)

    @service_method(
        path="worker_pool.add_workers",
        name="add_workers",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def add_workers(
        self,
        context: AuthedServiceContext,
        number: int,
        pool_id: UID | None = None,
        pool_name: str | None = None,
        registry_username: str | None = None,
        registry_password: str | None = None,
    ) -> list[ContainerSpawnStatus]:
        """Add workers to existing worker pool.

        Worker pool is fetched either using the unique pool id or pool name.

        Args:
            context (AuthedServiceContext): _description_
            number (int): number of workers to add
            pool_id (Optional[UID], optional): Unique UID of the pool. Defaults to None.
            pool_name (Optional[str], optional): Unique name of the pool. Defaults to None.

        Returns:
            List[ContainerSpawnStatus]: List of spawned workers with their status and error if any.
        """

        if number <= 0:
            raise SyftException(public_message=f"Invalid number of workers: {number}")

        # Extract pool using either using pool id or pool name
        if pool_id:
            worker_pool = self.stash.get_by_uid(
                credentials=context.credentials, uid=pool_id
            ).unwrap()
        elif pool_name:
            worker_pool = self.stash.get_by_name(
                credentials=context.credentials,
                pool_name=pool_name,
            ).unwrap()

        existing_worker_cnt = len(worker_pool.worker_list)

        worker_image = self.image_stash.get_by_uid(
            credentials=context.credentials,
            uid=worker_pool.image_id,
        ).unwrap()

        worker_stash = context.server.services.worker.stash

        # Add workers to given pool from the given image
        worker_list, container_statuses = _create_workers_in_pool(
            context=context,
            pool_name=worker_pool.name,
            existing_worker_cnt=existing_worker_cnt,
            worker_cnt=number,
            worker_image=worker_image,
            worker_stash=worker_stash,
            registry_username=registry_username,
            registry_password=registry_password,
        ).unwrap()

        worker_pool.worker_list += worker_list
        worker_pool.max_count = existing_worker_cnt + number

        self.stash.update(credentials=context.credentials, obj=worker_pool).unwrap()
        return container_statuses

    @service_method(
        path="worker_pool.scale",
        name="scale",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def scale(
        self,
        context: AuthedServiceContext,
        number: int,
        pool_id: UID | None = None,
        pool_name: str | None = None,
    ) -> SyftSuccess:
        """
        Scale the worker pool to the given number of workers in Kubernetes.
        Allows both scaling up and down the worker pool.
        """

        client_warning = ""

        if not IN_KUBERNETES:
            raise SyftException(
                public_message="Scaling is only supported in Kubernetes mode"
            )
        elif number < 0:
            # zero is a valid scale down
            raise SyftException(public_message=f"Invalid number of workers: {number}")

        worker_pool: Any = self._get_worker_pool(context, pool_id, pool_name).unwrap()
        current_worker_count = len(worker_pool.worker_list)

        if current_worker_count == number:
            return SyftSuccess(message=f"Worker pool already has {number} workers")
        elif number > current_worker_count:
            workers_to_add = number - current_worker_count
            self.add_workers(
                context=context,
                number=workers_to_add,
                pool_id=pool_id,
                pool_name=pool_name,
                # kube scaling doesn't require password as it replicates an existing deployment
                registry_username=None,
                registry_password=None,
            )
        else:
            # scale down at kubernetes control plane
            runner = KubernetesRunner()
            scale_kubernetes_pool(
                runner,
                pool_name=worker_pool.name,
                replicas=number,
            ).unwrap()

            # scale down removes the last "n" workers
            # workers to delete = len(workers) - number
            workers_to_delete = worker_pool.worker_list[
                -(current_worker_count - number) :
            ]

            worker_stash = context.server.services.worker.stash
            # delete linkedobj workers
            for worker in workers_to_delete:
                worker_stash.delete_by_uid(
                    credentials=context.credentials,
                    uid=worker.object_uid,
                ).unwrap()

            client_warning += "Scaling down workers doesn't kill the associated jobs. Please delete them manually."

            # update worker_pool
            worker_pool.max_count = number
            worker_pool.worker_list = worker_pool.worker_list[:number]
            self.stash.update(
                credentials=context.credentials,
                obj=worker_pool,
            ).unwrap(
                public_message=(
                    f"Pool {worker_pool.name} was scaled down, "
                    f"but failed to update the stash"
                )
            )

        return SyftSuccess(
            message=f"Worker pool scaled to {number} workers",
            client_warnings=[client_warning] if client_warning else [],
        )

    @service_method(
        path="worker_pool.filter_by_image_id",
        name="filter_by_image_id",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def filter_by_image_id(
        self, context: AuthedServiceContext, image_uid: UID
    ) -> list[WorkerPool]:
        return self.stash.get_by_image_uid(context.credentials, image_uid).unwrap()

    @service_method(
        path="worker_pool.get_by_name",
        name="get_by_name",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_by_name(
        self, context: AuthedServiceContext, pool_name: str
    ) -> list[WorkerPool]:
        return self.stash.get_by_name(context.credentials, pool_name).unwrap()

    @service_method(
        path="worker_pool.sync_pool_from_request",
        name="sync_pool_from_request",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def sync_pool_from_request(
        self,
        context: AuthedServiceContext,
        request: Request,
    ) -> Request:
        """Re-submit request from a different server"""

        num_of_changes = len(request.changes)
        pool_name, num_workers, config, image_uid, tag = None, None, None, None, None

        if num_of_changes > 2:
            raise SyftException(
                public_message=f"Invalid pool request object. Only pool request changes allowed. {request.changes}"
            )

        for change in request.changes:
            if isinstance(change, CreateCustomWorkerPoolChange):
                pool_name = change.pool_name
                num_workers = change.num_workers
                image_uid = change.image_uid
                pod_annotations = change.pod_annotations
                pod_labels = change.pod_labels
            elif isinstance(change, CreateCustomImageChange):  # type: ignore[unreachable]
                config = change.config
                tag = change.tag

        if config is None and image_uid is not None:
            return self.create_pool_request(
                context=context,
                pool_name=pool_name,
                num_workers=num_workers,
                image_uid=image_uid,
                pod_annotations=pod_annotations,
                pod_labels=pod_labels,
            )
        elif config is not None:
            return self.create_image_and_pool_request(  # type: ignore[unreachable]
                context=context,
                pool_name=pool_name,
                num_workers=num_workers,
                config=config,
                tag=tag,
                pod_annotations=pod_annotations,
                pod_labels=pod_labels,
            )
        else:
            raise SyftException(
                public_message=(
                    f"Invalid request object: invalid image uid or config in the request changes: "
                    f"{request.changes}"
                )
            )

    @service_method(
        path="worker_pool.delete",
        name="delete",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def delete(
        self,
        context: AuthedServiceContext,
        pool_id: UID | None = None,
        pool_name: str | None = None,
    ) -> SyftSuccess:
        worker_pool = self._get_worker_pool(
            context, pool_id=pool_id, pool_name=pool_name
        ).unwrap(public_message=f"Failed to get WorkerPool: {pool_id or pool_name}")

        uid = worker_pool.id

        self.purge_workers(context=context, pool_id=pool_id, pool_name=pool_name)

        self.stash.delete_by_uid(credentials=context.credentials, uid=uid).unwrap(
            public_message=f"Failed to delete WorkerPool: {worker_pool.name} from stash"
        )

        return SyftSuccess(message=f"Successfully deleted worker pool with id {uid}")

    @service_method(
        path="worker_pool.purge_workers",
        name="purge_workers",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def purge_workers(
        self,
        context: AuthedServiceContext,
        pool_id: UID | None = None,
        pool_name: str | None = None,
    ) -> SyftSuccess:
        worker_pool = self._get_worker_pool(
            context, pool_id=pool_id, pool_name=pool_name
        ).unwrap(public_message=f"Failed to get WorkerPool: {pool_id or pool_name}")

        uid = worker_pool.id

        # relative
        from ..queue.queue_stash import Status

        queue_items = context.server.services.queue.stash._get_by_worker_pool(
            credentials=context.credentials,
            worker_pool=LinkedObject.from_obj(
                obj=worker_pool,
                service_type=self.__class__,
                server_uid=context.server.id,
            ),
        ).unwrap(
            public_message=f"Failed to get queue items mapped to WorkerPool: {worker_pool.name}"
        )

        items_to_interrupt = (
            item
            for item in queue_items
            if item.status in (Status.CREATED, Status.PROCESSING)
        )

        for item in items_to_interrupt:
            item.status = Status.INTERRUPTED
            context.server.services.queue.stash.update(
                credentials=context.credentials,
                obj=item,
            ).unwrap()

        if IN_KUBERNETES:
            # Scale the workers to zero
            runner = KubernetesRunner()
            if runner.exists(worker_pool.name):
                self.scale(context=context, number=0, pool_id=uid)
                runner.delete_pool(pool_name=worker_pool.name)
        else:
            workers = (
                worker.resolve_with_context(context=context).unwrap()
                for worker in worker_pool.worker_list
            )

            worker_ids = []
            for worker in workers:
                worker_ids.append(worker.id)

            for id_ in worker_ids:
                context.server.services.worker.delete(
                    context=context, uid=id_, force=True
                )

        worker_pool.max_count = 0
        worker_pool.worker_list = []
        self.stash.update(
            credentials=context.credentials,
            obj=worker_pool,
        ).unwrap(
            public_message=(
                f"Pool {worker_pool.name} was purged, "
                f"but failed to update the stash"
            )
        )

        return SyftSuccess(message=f"Successfully Purged worker pool with id {uid}")

    @as_result(StashException, SyftException)
    def _get_worker_pool(
        self,
        context: AuthedServiceContext,
        pool_id: UID | None = None,
        pool_name: str | None = None,
    ) -> WorkerPool:
        if pool_id:
            worker_pool = self.stash.get_by_uid(
                credentials=context.credentials,
                uid=pool_id,
            ).unwrap()
        else:
            worker_pool = self.stash.get_by_name(
                credentials=context.credentials,
                pool_name=pool_name,
            ).unwrap()

        return worker_pool


@as_result(SyftException)
def _create_workers_in_pool(
    context: AuthedServiceContext,
    pool_name: str,
    existing_worker_cnt: int,
    worker_cnt: int,
    worker_image: SyftWorkerImage,
    worker_stash: WorkerStash,
    registry_username: str | None = None,
    registry_password: str | None = None,
    pod_annotations: dict[str, str] | None = None,
    pod_labels: dict[str, str] | None = None,
) -> tuple[list[LinkedObject], list[ContainerSpawnStatus]]:
    queue_port = context.server.queue_config.client_config.queue_port

    # Check if workers needs to be run in memory or as containers
    start_workers_in_memory = context.server.in_memory_workers

    if start_workers_in_memory:
        # Run in-memory workers in threads
        container_statuses: list[ContainerSpawnStatus] = run_workers_in_threads(
            server=context.server,
            pool_name=pool_name,
            start_idx=existing_worker_cnt,
            number=worker_cnt + existing_worker_cnt,
        )
    else:
        registry_host = (
            worker_image.image_identifier.registry_host
            if worker_image.image_identifier is not None
            else None
        )
        container_statuses = run_containers(
            pool_name=pool_name,
            worker_image=worker_image,
            start_idx=existing_worker_cnt,
            number=worker_cnt + existing_worker_cnt,
            orchestration=get_orchestration_type(),
            queue_port=queue_port,
            dev_mode=context.server.dev_mode,
            registry_username=registry_username,
            registry_password=registry_password,
            reg_url=registry_host,
            pod_annotations=pod_annotations,
            pod_labels=pod_labels,
        ).unwrap()

    linked_worker_list = []

    for container_status in container_statuses:
        worker = container_status.worker
        if worker is None:
            continue

        server = context.server

        try:
            obj = worker_stash.set(
                credentials=context.credentials,
                obj=worker,
            ).unwrap()

            worker_obj = LinkedObject.from_obj(
                obj=obj,
                service_type=WorkerService,
                server_uid=server.id,
            )

            linked_worker_list.append(worker_obj)
        except SyftException as exc:
            container_status.error = exc.public_message
    return linked_worker_list, container_statuses


TYPE_TO_SERVICE[WorkerPool] = SyftWorkerPoolService
SERVICE_TO_TYPES[SyftWorkerPoolService] = WorkerPool



# stdlib
from pathlib import Path
import tempfile
import uuid

# third party
from pydantic import Field
import sqlalchemy as sa

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...types.uid import UID
from .db import DBConfig
from .db import DBManager


@serializable(canonical_name="SQLiteDBConfig", version=1)
class SQLiteDBConfig(DBConfig):
    filename: str = Field(default_factory=lambda: f"{uuid.uuid4()}.db")
    path: Path = Field(default_factory=lambda: Path(tempfile.gettempdir()))

    @property
    def connection_string(self) -> str:
        """
        NOTE in-memory sqlite is not shared between connections, so:
        - using 2 workers (high/low) will not share a db
        - re-using a connection (e.g. for a Job worker) will not share a db
        """
        if self.path == Path("."):
            # Use in-memory database, only for unittests
            return "sqlite://"
        filepath = self.path / self.filename
        return f"sqlite:///{filepath.resolve()}"


class SQLiteDBManager(DBManager[SQLiteDBConfig]):
    def update_settings(self) -> None:
        connection = self.engine.connect()
        connection.execute(sa.text("PRAGMA journal_mode = WAL"))
        connection.execute(sa.text("PRAGMA busy_timeout = 5000"))
        connection.execute(sa.text("PRAGMA temp_store = 2"))
        connection.execute(sa.text("PRAGMA synchronous = 1"))

    @classmethod
    def random(
        cls,
        *,
        config: SQLiteDBConfig | None = None,
        server_uid: UID | None = None,
        root_verify_key: SyftVerifyKey | None = None,
    ) -> "SQLiteDBManager":
        root_verify_key = root_verify_key or SyftSigningKey.generate().verify_key
        server_uid = server_uid or UID()
        config = config or SQLiteDBConfig()
        return SQLiteDBManager(
            config=config,
            server_uid=server_uid,
            root_verify_key=root_verify_key,
        )



# syft absolute
import syft as sy

# relative
from ...client.client import SyftClient
from ...service.response import SyftSuccess
from ...service.worker.worker_image import SyftWorkerImage
from ...types.uid import UID


def build_and_launch_worker_pool_from_docker_str(
    environment: str,
    client: sy.DatasiteClient,
    worker_pool_name: str,
    custom_pool_pod_annotations: dict,
    custom_pool_pod_labels: dict,
    worker_dockerfile: str,
    external_registry: str,
    docker_tag: str,
    scale_to: int,
):
    result = client.api.services.image_registry.add(external_registry)
    assert "success" in result.message  # nosec: B101

    # For some reason, when using k9s, result.value is empty so can't use the below line
    # local_registry = result.value
    local_registry = client.api.services.image_registry[0]

    docker_config = sy.DockerWorkerConfig(dockerfile=worker_dockerfile)
    assert docker_config.dockerfile == worker_dockerfile  # nosec: B101
    submit_result = client.api.services.worker_image.submit(worker_config=docker_config)
    print(submit_result.message)
    assert "success" in submit_result.message  # nosec: B101

    worker_image = submit_result.value

    if environment == "remote":
        docker_build_result = client.api.services.worker_image.build(
            image_uid=worker_image.id,
            tag=docker_tag,
            registry_uid=local_registry.id,
        )
        print(docker_build_result)

    if environment == "remote":
        push_result = client.api.services.worker_image.push(worker_image.id)
        print(push_result)

    result = client.api.services.worker_pool.launch(
        pool_name=worker_pool_name,
        image_uid=worker_image.id,
        num_workers=1,
        pod_annotations=custom_pool_pod_annotations,
        pod_labels=custom_pool_pod_labels,
    )
    print(result)
    # assert 'success' in str(result.message)

    # scale_to > 1 is valid for scale up
    # scale_to = 0 is valid for removing all pods
    # scale_to < 0 should return error from server
    if environment == "remote" and scale_to != 1:
        result = client.worker_pools.scale(number=scale_to, pool_name=worker_pool_name)
        print(result)


def launch_worker_pool_from_docker_tag_and_registry(
    environment: str,
    client: sy.DatasiteClient,
    worker_pool_name: str,
    custom_pool_pod_annotations: dict,
    custom_pool_pod_labels: dict,
    docker_tag: str,
    external_registry: str,
    scale_to: int = 1,
):
    res = client.api.services.image_registry.add(external_registry)
    assert "success" in res.message  # nosec: B101
    docker_config = sy.PrebuiltWorkerConfig(tag=docker_tag)
    image_result = client.api.services.worker_image.submit(worker_config=docker_config)
    assert "success" in res.message  # nosec: B101
    worker_image = image_result.value

    launch_result = client.api.services.worker_pool.launch(
        pool_name=worker_pool_name,
        image_uid=worker_image.id,
        num_workers=1,
        pod_annotations=custom_pool_pod_annotations,
        pod_labels=custom_pool_pod_labels,
    )
    if environment == "remote" and scale_to > 1:
        result = client.worker_pools.scale(number=scale_to, pool_name=worker_pool_name)
        print(result)

    return launch_result


def prune_worker_pool_and_images(client: SyftClient) -> None:
    for pool in client.worker_pools.get_all():
        client.worker_pools.delete(pool.id)

    for image in client.images.get_all():
        client.images.remove(image.id)


def build_and_push_image(
    client: SyftClient,
    image: SyftWorkerImage,
    tag: str,
    registry_uid: UID | None = None,
    reg_username: str | None = None,
    reg_password: str | None = None,
    force_build: bool = False,
) -> None:
    """Build and push the image to the given registry."""
    if image.is_prebuilt:
        return

    build_result = client.api.services.worker_image.build(
        image_uid=image.id, registry_uid=registry_uid, tag=tag, force_build=force_build
    )
    print(build_result.message)

    if isinstance(build_result, SyftSuccess):
        push_result = client.api.services.worker_image.push(
            image.id,
            username=reg_username,
            password=reg_password,
        )
        assert isinstance(push_result, SyftSuccess)  # nosec: B101
        print(push_result.message)



# stdlib
from collections.abc import Callable
import functools
from typing import Any
from typing import Final
from typing import Generic
from typing import Literal
from typing import NoReturn
from typing import ParamSpec
from typing import TypeAlias
from typing import TypeVar

# relative
from .errors import SyftException
from .errors import exclude_from_traceback
from .errors import process_traceback

T = TypeVar("T", covariant=True)
E = TypeVar("E", covariant=True, bound=BaseException)
BE = TypeVar("BE", bound=BaseException)
P = ParamSpec("P")


class Ok(Generic[T]):
    __slots__ = ("value",)
    __match_args__ = ("ok_value",)

    def __init__(self, value: T):
        self.value = value

    def __repr__(self) -> str:
        return f"Ok({self.value})"

    @property
    def ok_value(self) -> T:
        return self.value

    def err(self) -> None:
        return None

    def ok(self) -> T:
        return self.value

    def is_err(self) -> Literal[False]:
        return False

    def is_ok(self) -> Literal[True]:
        return True

    def unwrap(self, *args: Any, **kwargs: Any) -> T:
        return self.value


class Err(Generic[E]):
    __slots__ = ("value",)
    __match_args__ = ("error_value",)

    def __init__(self, value: E):
        self.value = value

    def __repr__(self) -> str:
        return f"Err({self.value})"

    @property
    def error_value(self) -> E:
        return self.value

    def err(self) -> E:
        return self.value

    def ok(self) -> None:
        return None

    def is_err(self) -> Literal[True]:
        return True

    def is_ok(self) -> Literal[False]:
        return False

    @exclude_from_traceback
    def unwrap(
        self, public_message: str | None = None, private_message: str | None = None
    ) -> NoReturn:
        if isinstance(self.value, SyftException):
            if public_message is not None:
                self.value.public_message = public_message
            if private_message is not None:
                self.value._private_message = private_message
        if isinstance(self.value, BaseException):
            raise self.value
        raise TypeError("Error is not a BaseException")


OkErr: Final = (Ok, Err)
Result: TypeAlias = Ok[T] | Err[E]


def as_result(
    *exceptions: type[BE], convert_to_syft_exception: bool = False
) -> Callable[[Callable[P, T]], Callable[P, Result[T, BE]]]:
    if not exceptions or not all(
        issubclass(exception, BaseException) for exception in exceptions
    ):
        raise TypeError("The as_result() decorator only accepts exceptions")

    class _AsResultError(Exception): ...

    def decorator(func: Callable[P, T]) -> Callable[P, Result[T, BE]]:
        @exclude_from_traceback
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[T, BE]:
            try:
                output = func(*args, **kwargs)
                if isinstance(output, Ok) or isinstance(output, Err):
                    raise _AsResultError(
                        f"Functions decorated with `as_result` should not return Result.\n"
                        f"Did you forget to unwrap() the result in {func.__name__}?\n"
                        f"result: {output}"
                    )
                return Ok(output)
            except exceptions as exc:
                if convert_to_syft_exception and not isinstance(exc, SyftException):
                    exc = SyftException.from_exception(exc)  # type: ignore
                exc = process_traceback(exc)
                return Err(exc)

        return wrapper

    return decorator
